;
; Copyright (c) 2012, Brandon Falk <bfalk@brandonfa.lk>
; All rights reserved.
;
; Redistribution and use in source and binary forms, with or without
; modification, are permitted provided that the following conditions are met:
;
; 	1. Redistributions of source code must retain the above copyright notice,
; 	this list of conditions and the following disclaimer.
;
; 	2. Redistributions in binary form must reproduce the above copyright
; 	notice, this list of conditions and the following disclaimer in the
; 	documentation and/or other materials provided with the distribution.
;
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
; "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
; LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
; PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
; OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
; EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
; PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
; WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
; ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
;

[bits 64]

section .text
	global WinMainCRTStartup
	extern CreateThread
	extern WaitForMultipleObjects
	extern ExitProcess

%define threads 12
%define hashes  1000000000

%macro cthrwerr 0
	push thrid        ; thread ID holder
	push 0            ; flags (run right away)
	xor  r9,  r9      ; parameters
	mov  r8,  _genmd5 ; start address
	xor rdx,  rdx     ; stack size (default)
	xor rcx,  rcx     ; security attribs
	call CreateThread

	mov qword [r13], rax
	add r13, 8

	cmp rax, 0
	je  .die
%endmacro

%macro rots1 0
	mov r10d,  edx ; temp = d
	mov  edx,  ecx ; d    = c
	mov  ecx,  ebx ; c    = b
%endmacro

%macro shifto 3
	; b += rol((a + f + k[i] + w[g]) , r[i])
	add  eax,  r8d
	add  eax,  %1
	add  eax,  dword [str + %2]
	rol  eax,  %3
	add  ebx,  eax
	mov  eax, r10d ; a = temp
%endmacro

%macro stg1 3
	mov  r8d,  ebx ; r8d = b
	and  r8d,  ecx ; r8d = (b & c)
	mov  r9d,  ebx ; r9d = b
	not  r9d       ; r9d = ~b
	and  r9d,  edx ; r9d = ((~b) & d)
	or   r8d,  r9d ; r8d = ((b & c) | ((~b) & d))

	rots1
	shifto %1, %2, %3
%endmacro

%macro stg2 3
	mov  r8d,  edx ; r8d = d
	and  r8d,  ebx ; r8d = (d & b)
	mov  r9d,  edx ; r9d = d
	not  r9d       ; r9d = ~d
	and  r9d,  ecx ; r9d = ((~d) & c)
	or   r8d,  r9d ; r8d = ((d & b) | ((~d) & c))

	rots1
	shifto %1, %2, %3
%endmacro

%macro stg3 3
	mov  r8d, ebx ; r8d = c
	xor  r8d, ecx ; r8d = (b ^ c)
	xor  r8d, edx ; r8d = ((b ^ c) ^ d)

	rots1
	shifto %1, %2, %3
%endmacro

%macro stg4 3
	mov  r8d, edx ; r8d = d
	not  r8d      ; r8d = ~d
	or   r8d, ebx ; r8d = (b | (~d))
	xor  r8d, ecx ; r8d = (c ^ (b | (~d)))

	rots1
	shifto %1, %2, %3
%endmacro

_genmd5:
	mov r15, (hashes / threads)

.lewp:
	mov eax, 0x67452301
	mov ebx, 0xEFCDAB89
	mov ecx, 0x98BADCFE
	mov edx, 0x10325476

	; Round 1 is static
	mov  r8d,  ecx ; f
	rots1
	shifto 0xd76aa478, 0, 7

	stg1 0xe8c7b756, 4, 12
	stg1 0x242070db, 8, 17
	stg1 0xc1bdceee, 12, 22
	stg1 0xf57c0faf, 16, 7
	stg1 0x4787c62a, 20, 12
	stg1 0xa8304613, 24, 17
	stg1 0xfd469501, 28, 22
	stg1 0x698098d8, 32, 7
	stg1 0x8b44f7af, 36, 12
	stg1 0xffff5bb1, 40, 17
	stg1 0x895cd7be, 44, 22
	stg1 0x6b901122, 48, 7
	stg1 0xfd987193, 52, 12
	stg1 0xa679438e, 56, 17
	stg1 0x49b40821, 60, 22

	stg2 0xf61e2562, 4, 5
	stg2 0xc040b340, 24, 9
	stg2 0x265e5a51, 44, 14
	stg2 0xe9b6c7aa, 0, 20
	stg2 0xd62f105d, 20, 5
	stg2 0x02441453, 40, 9
	stg2 0xd8a1e681, 60, 14
	stg2 0xe7d3fbc8, 16, 20
	stg2 0x21e1cde6, 36, 5
	stg2 0xc33707d6, 56, 9
	stg2 0xf4d50d87, 12, 14
	stg2 0x455a14ed, 32, 20
	stg2 0xa9e3e905, 52, 5
	stg2 0xfcefa3f8, 8, 9
	stg2 0x676f02d9, 28, 14
	stg2 0x8d2a4c8a, 48, 20

	stg3 0xfffa3942, 20, 4
	stg3 0x8771f681, 32, 11
	stg3 0x6d9d6122, 44, 16
	stg3 0xfde5380c, 56, 23
	stg3 0xa4beea44, 4, 4
	stg3 0x4bdecfa9, 16, 11
	stg3 0xf6bb4b60, 28, 16
	stg3 0xbebfbc70, 40, 23
	stg3 0x289b7ec6, 52, 4
	stg3 0xeaa127fa, 0, 11
	stg3 0xd4ef3085, 12, 16
	stg3 0x04881d05, 24, 23
	stg3 0xd9d4d039, 36, 4
	stg3 0xe6db99e5, 48, 11
	stg3 0x1fa27cf8, 60, 16
	stg3 0xc4ac5665, 8, 23

	stg4 0xf4292244, 0, 6
	stg4 0x432aff97, 28, 10
	stg4 0xab9423a7, 56, 15
	stg4 0xfc93a039, 20, 21
	stg4 0x655b59c3, 48, 6
	stg4 0x8f0ccc92, 12, 10
	stg4 0xffeff47d, 40, 15
	stg4 0x85845dd1, 4, 21
	stg4 0x6fa87e4f, 32, 6
	stg4 0xfe2ce6e0, 60, 10
	stg4 0xa3014314, 24, 15
	stg4 0x4e0811a1, 52, 21
	stg4 0xf7537e82, 16, 6
	stg4 0xbd3af235, 44, 10
	stg4 0x2ad7d2bb, 8, 15
	stg4 0xeb86d391, 36, 21

	add  eax, 0x67452301
	add  ebx, 0xEFCDAB89
	add  ecx, 0x98BADCFE
	add  edx, 0x10325476

	dec r15
	jne .lewp

	; Hashing done

	ret

WinMainCRTStartup:
	sub rsp, 0x28

	mov r13, hdls
	%rep threads
	cthrwerr
	%endrep

	xor r10, r10
	not r10
	xor  r9, r9
	not  r9
	mov rdx, hdls
	mov rcx, threads
	call WaitForMultipleObjects

.die:
	add rsp, 0x28
	call ExitProcess
	ret

section .data

thrid:
	dq 0
hdls:
	times threads dq 0

str:
	db 'The quick brown fox jumps over the lazy dog'
strend:
	db 0x80
	times (56 - ($ - str)) db 0
	dq ((strend - str) * 8)

